import torch 
import torch.nn as nn
import torch.nn.functional as F
import sys
sys.path.append("./torchdp/")

#from torchdp.autograd_grad_sample import add_hooks, remove_hooks
# from torchdp.autograd_grad_sample import add_hooks, remove_hooks
from torch.autograd import Variable
import numpy as np

# Loss functions
def loss_coteaching(y_1, y_2, t, forget_rate, ind, noise_or_not):
    noise_or_not = torch.Tensor(noise_or_not).cuda()
    ind = torch.Tensor(ind).cuda().long()
    loss_1 = F.cross_entropy(y_1, t, reduction='none')
    ind_1_sorted = torch.argsort(loss_1).cuda()
    loss_1_sorted = loss_1[ind_1_sorted]

    loss_2 = F.cross_entropy(y_2, t, reduction='none')
    ind_2_sorted = torch.argsort(loss_2).cuda()
    loss_2_sorted = loss_2[ind_2_sorted]

    remember_rate = 1 - forget_rate
    num_remember = int(remember_rate * len(loss_1_sorted))

    pure_ratio_1 = torch.sum(noise_or_not[ind[ind_1_sorted[:num_remember]]])/float(num_remember)
    pure_ratio_2 = torch.sum(noise_or_not[ind[ind_2_sorted[:num_remember]]])/float(num_remember)

    ind_1_update=ind_1_sorted[:num_remember]
    ind_2_update=ind_2_sorted[:num_remember]
    # exchange
    loss_1_update = F.cross_entropy(y_1[ind_2_update], t[ind_2_update])
    loss_2_update = F.cross_entropy(y_2[ind_1_update], t[ind_1_update])

    return torch.sum(loss_1_update)/num_remember, torch.sum(loss_2_update)/num_remember, pure_ratio_1, pure_ratio_2

def loss_spl(y_1, y_2, t, forget_rate, ind, noise_or_not):
    noise_or_not = torch.Tensor(noise_or_not).cuda()
    ind = torch.Tensor(ind).cuda().long()
    loss_1 = F.cross_entropy(y_1, t, reduction='none')
    ind_1_sorted = torch.argsort(loss_1).cuda()
    loss_1_sorted = loss_1[ind_1_sorted]

    loss_2 = F.cross_entropy(y_2, t, reduction='none')
    ind_2_sorted = torch.argsort(loss_2).cuda()
    loss_2_sorted = loss_2[ind_2_sorted]

    remember_rate = 1 - forget_rate
    num_remember = int(remember_rate * len(loss_1_sorted))

    pure_ratio_1 = torch.sum(noise_or_not[ind[ind_1_sorted[:num_remember]]])/float(num_remember)
    pure_ratio_2 = torch.sum(noise_or_not[ind[ind_2_sorted[:num_remember]]])/float(num_remember)

    ind_1_update=ind_1_sorted[:num_remember]
    ind_2_update=ind_2_sorted[:num_remember]
    # exchange
    loss_1_update = F.cross_entropy(y_1[ind_1_update], t[ind_1_update])
    loss_2_update = F.cross_entropy(y_2[ind_2_update], t[ind_2_update])

    return torch.sum(loss_1_update)/num_remember, torch.sum(loss_2_update)/num_remember, pure_ratio_1, pure_ratio_2



def loss_signsgd(x, model1, model2, t):
    # you should assert the model1 and model2 are added hooks
    add_hooks(model1)
    add_hooks(model2)
    y_1 = model1(x)
    y_2 = model2(x)


    loss_1 = F.cross_entropy(y_1, t, reduction='none')
    loss_1_bp = torch.sum(loss_1)
    loss_1_bp.backward()

    loss_2 = F.cross_entropy(y_2, t, reduction='none')
    loss_2_bp = torch.sum(loss_2)
    loss_2_bp.backward()
    # get individual gradient
    with torch.no_grad():
        n = y_1.shape[0]
        for p in model1.parameters():
            grad_layer = (p.grad_sample.detach()).squeeze()
            grad_layer = (((grad_layer > 0)).sum(axis=0) > 0).float()
            grad_layer = grad_layer * 2 - 1
            p.grad = grad_layer.reshape(p.grad.shape).clone()
            del p.grad_sample

        for p in model2.parameters():
            grad_layer = (p.grad_sample.detach()).squeeze()
            grad_layer = (((grad_layer > 0)).sum(axis=0) > 0).float()
            grad_layer = grad_layer * 2 - 1
            p.grad = grad_layer.reshape(p.grad.shape).clone()
            del p.grad_sample


    remove_hooks(model1)
    remove_hooks(model2)
    # exchange



def loss_CO_IGFilter(x, model1, model2, t, forget_rate, ind, noise_or_not):
    # you should assert the model1 and model2 are added hooks
    noise_or_not = torch.Tensor(noise_or_not).cuda()
    ind = torch.Tensor(ind).cuda().long()
    add_hooks(model1)
    add_hooks(model2)
    y_1 = model1(x)
    y_2 = model2(x)


    loss_1 = F.cross_entropy(y_1, t, reduction='none')
    loss_1_bp = torch.sum(loss_1)
    loss_1_bp.backward()

    loss_2 = F.cross_entropy(y_2, t, reduction='none')
    loss_2_bp = torch.sum(loss_2)
    loss_2_bp.backward()
    # get individual gradient
    with torch.no_grad():
        n = y_1.shape[0]
        # computed_sample_grads_1 = []
        computed_grad_norm_1 = torch.zeros(n).cuda()

        for p in model1.parameters():
            grad_layer = torch.flatten(p.grad_sample.detach(), start_dim=1).squeeze()

            # grad_layer = grad_layer.clamp(-0.5, 0.5)
            # computed_sample_grads_1.append(grad_layer)
            # print(grad_layer.shape)
            # print((grad_layer.mean(dim=0)).shape)
            # grad_layer = (grad_layer - grad_layer.mean(dim=0))
            # print(grad_layer)

            # cov_mat = (grad_layer.mm(grad_layer.transpose(1, 0)))
            # cov_mat = torch.sum(cov_mat > 0, dim=1)

            del p.grad_sample
            # print(cov_mat)
            computed_grad_norm_1 = computed_grad_norm_1 + torch.sum(grad_layer**2, dim=1)
            # computed_grad_norm_1 = computed_grad_norm_1 + cov_mat
            # print(computed_grad_norm_1.shape)
            # computed_sample_grads_1.append(torch.flatten(p.grad_sample.detach(), start_dim=1).squeeze())
        # computed_sample_grads_1 = torch.cat(computed_sample_grads_1, axis=1)
        # computed_sample_grads_1 = computed_sample_grads_1 > 0
        # computed_sample_grads_1 = computed_sample_grads_1.sum(axis=1)
        # print(computed_grad_norm_1)
        computed_grad_norm_2 = torch.zeros(n).cuda()
        # computed_sample_grads_2 = []
        for p in model2.parameters():
            grad_layer = torch.flatten(p.grad_sample.detach(), start_dim=1).squeeze()
            # grad_layer = (grad_layer - grad_layer.mean(dim=0))
            # computed_sample_grads_2.append(grad_layer)
            # cov_mat = torch.sum(cov_mat > 0, dim=1)
            # grad_layer = (grad_layer - grad_layer.mean(dim=0))
            computed_grad_norm_2 = computed_grad_norm_2 + torch.sum(grad_layer ** 2, dim=1)
            # computed_grad_norm_2 = computed_grad_norm_2 + cov_mat
        # computed_sample_grads_2 = torch.cat(computed_sample_grads_2, axis=1)
            del p.grad_sample
        # computed_sample_grads_2 = torch.cat(computed_sample_grads_2, axis=1)

        _, ind_1_sorted = torch.sort(computed_grad_norm_1)
        _, ind_2_sorted = torch.sort(computed_grad_norm_2)

    # ind_1_sorted = np.argsort(loss_1.data).cuda()

    loss_1_sorted = loss_1[ind_1_sorted.squeeze()]

    # loss_2 = F.cross_entropy(y_2, t, reduce = False)
    # ind_2_sorted = np.argsort(loss_2.data).cuda()
    loss_2_sorted = loss_2[ind_2_sorted.squeeze()]

    remember_rate = 1 - forget_rate
    num_remember = int(remember_rate * len(loss_1_sorted))
    num_start = int((1-remember_rate) * len(loss_1_sorted))

    pure_ratio_1 = torch.sum(noise_or_not[ind[ind_1_sorted[:num_remember]]])/float(num_remember)
    pure_ratio_2 = torch.sum(noise_or_not[ind[ind_2_sorted[:num_remember]]])/float(num_remember)

    ind_1_update=ind_1_sorted[: num_remember]
    ind_2_update=ind_2_sorted[: num_remember]

    remove_hooks(model1)
    remove_hooks(model2)
    # exchange
    model1.zero_grad()
    model2.zero_grad()
    y_1 = model1(x[ind_2_update, :, :, :])
    y_2 = model2(x[ind_1_update, :, :, :])
    loss_1_update = F.cross_entropy(y_1, t[ind_2_update])
    loss_2_update = F.cross_entropy(y_2, t[ind_1_update])

    # y_1 = model1(x[ind_2_update, :, :, :])
    # y_2 = model2(x[ind_1_update, :, :, :])
    #
    # loss_1 = F.cross_entropy(y_1, t[ind_2_update], reduction='none')
    # loss_1_bp = torch.sum(loss_1)
    # # loss_1_bp.backward()
    #
    # loss_2 = F.cross_entropy(y_2, t[ind_1_update], reduction='none')
    # loss_2_bp = torch.sum(loss_2)
    # loss_2_bp.backward()
    return torch.sum(loss_1_update)/num_remember, torch.sum(loss_2_update)/num_remember, pure_ratio_1, pure_ratio_2




# def loss_CO_DoubleRobust(x, model1, model2, t, forget_rate, ind, noise_or_not, which_model_grad = 1):
#     # you should assert the model1 and model2 are added hooks
#     noise_or_not = torch.Tensor(noise_or_not).cuda()
#     ind = torch.Tensor(ind).cuda().long()
#     if which_model_grad == 1:
#         add_hooks(model1)
#     elif which_model_grad == 2:
#         add_hooks(model2)
#     else:
#         raise Exception("which_model_grad should be 1 or 2")
#
#     y_1 = model1(x)
#     y_2 = model2(x)
#
#     loss_1 = F.cross_entropy(y_1, t, reduction='none')
#     loss_1_bp = torch.sum(loss_1)
#     loss_1_bp.backward()
#
#     loss_2 = F.cross_entropy(y_2, t, reduction='none')
#     loss_2_bp = torch.sum(loss_2)
#     loss_2_bp.backward()
#     # get individual gradient
#     with torch.no_grad():
#         n = y_1.shape[0]
#         if which_model_grad == 1:
#             computed_grad_norm_1 = torch.zeros(n).cuda()
#             for p in model1.parameters():
#                 grad_layer = torch.flatten(p.grad_sample.detach(), start_dim=1).squeeze()
#                 # grad_layer = (grad_layer - grad_layer.mean(dim=0))
#                 del p.grad_sample
#                 computed_grad_norm_1 = computed_grad_norm_1 + torch.sum(grad_layer**2, dim=1)
#
#             _, ind_1_sorted = torch.sort(computed_grad_norm_1)
#             _, ind_2_sorted = torch.sort(loss_2)
#         else:
#             computed_grad_norm_2 = torch.zeros(n).cuda()
#             for p in model2.parameters():
#                 grad_layer = torch.flatten(p.grad_sample.detach(), start_dim=1).squeeze()
#                 # grad_layer = (grad_layer - grad_layer.mean(dim=0))
#                 del p.grad_sample
#                 computed_grad_norm_2 = computed_grad_norm_2 + torch.sum(grad_layer ** 2, dim=1)
#
#             _, ind_1_sorted = torch.sort(loss_1)
#             _, ind_2_sorted = torch.sort(computed_grad_norm_2)
#
#     loss_1_sorted = loss_1[ind_1_sorted.squeeze()]
#     loss_2_sorted = loss_2[ind_2_sorted.squeeze()]
#
#     remember_rate = 1 - forget_rate
#     num_remember = int(remember_rate * len(loss_1_sorted))
#     num_start = int((1-remember_rate) * len(loss_1_sorted))
#
#     pure_ratio_1 = torch.sum(noise_or_not[ind[ind_1_sorted[:num_remember]]])/float(num_remember)
#     pure_ratio_2 = torch.sum(noise_or_not[ind[ind_2_sorted[:num_remember]]])/float(num_remember)
#
#     ind_1_update=ind_1_sorted[: num_remember]
#     ind_2_update=ind_2_sorted[: num_remember]
#
#     if which_model_grad == 1:
#         remove_hooks(model1)
#     else:
#         remove_hooks(model2)
#     # exchange
#
#     selected_index = torch.from_numpy(np.intersect1d(ind_1_update.cpu().numpy(),ind_2_update.cpu().numpy()))
#     model1.zero_grad()
#     model2.zero_grad()
#     y_1 = model1(x[selected_index, :, :, :])
#     y_2 = model2(x[selected_index, :, :, :])
#     loss_1_update = F.cross_entropy(y_1, t[selected_index])
#     loss_2_update = F.cross_entropy(y_2, t[selected_index])
#     return torch.sum(loss_1_update)/num_remember, torch.sum(loss_2_update)/num_remember, pure_ratio_1, pure_ratio_2

def loss_IGFilter_Classification(x, model1, model2, t, forget_rate, ind, noise_or_not, which_model_grad = 1, num_classes=10):
    noise_or_not = torch.Tensor(noise_or_not).cuda()
    ind = torch.Tensor(ind).cuda().long()

    y_1 = model1(x)
    y_2 = model2(x)

    loss_1 = F.cross_entropy(y_1, t, reduction='none')
    loss_1_bp = torch.sum(loss_1)
    # loss_1_bp.backward()
    loss_1 = F.mse_loss(F.softmax(y_1, dim=1), torch.nn.functional.one_hot(t, num_classes=num_classes), reduction='none')
    loss_1 = loss_1.sum(dim=1)
    loss_2 = F.cross_entropy(y_2, t, reduction='none')
    loss_2_bp = torch.sum(loss_2)
    # loss_2_bp.backward()
    loss_2 = F.mse_loss(F.softmax(y_2, dim=1), torch.nn.functional.one_hot(t, num_classes=num_classes), reduction='none')
    loss_2 = loss_2.sum(dim=1)
    # get individual gradient
    with torch.no_grad():
        n = y_1.shape[0]
        remember_rate = 1 - forget_rate
        num_remember = int(remember_rate * n)
        num_start = int((1 - remember_rate) * n)

        _, ind_1_sorted = torch.sort(loss_1)
        ind_1_update = ind_1_sorted[:num_remember]

        _, ind_2_sorted = torch.sort(loss_2)
        ind_2_update = ind_2_sorted[:num_remember]


    pure_ratio_1 = torch.sum(noise_or_not[ind[ind_1_sorted[:num_remember]]])/float(num_remember)
    pure_ratio_2 = torch.sum(noise_or_not[ind[ind_2_sorted[:num_remember]]])/float(num_remember)

    # exchange

    model1.zero_grad()
    model2.zero_grad()
    y_1 = model1(x[ind_2_update, :, :, :])
    y_2 = model2(x[ind_1_update, :, :, :])
    loss_1_update = F.cross_entropy(y_1, t[ind_2_update])
    loss_2_update = F.cross_entropy(y_2, t[ind_1_update])
    return torch.sum(loss_1_update) / num_remember, torch.sum(loss_2_update) / num_remember, pure_ratio_1, pure_ratio_2


def loss_spl_IGFilter_Classification(x, model1, model2, t, forget_rate, ind, noise_or_not, which_model_grad = 1, num_classes=10):
    noise_or_not = torch.Tensor(noise_or_not).cuda()
    ind = torch.Tensor(ind).cuda().long()
    
    y_1 = model1(x)
    y_2 = model2(x)
    
    loss_1 = F.cross_entropy(y_1, t, reduction='none')
    loss_1_bp = torch.sum(loss_1)
    # loss_1_bp.backward()
    loss_1 = F.mse_loss(F.softmax(y_1, dim=1), torch.nn.functional.one_hot(t, num_classes=num_classes), reduction='none')
    loss_1 = loss_1.sum(dim=1)
    loss_2 = F.cross_entropy(y_2, t, reduction='none')
    loss_2_bp = torch.sum(loss_2)
    # loss_2_bp.backward()
    loss_2 = F.mse_loss(F.softmax(y_2, dim=1), torch.nn.functional.one_hot(t, num_classes=num_classes), reduction='none')
    loss_2 = loss_2.sum(dim=1)
    # get individual gradient
    with torch.no_grad():
        n = y_1.shape[0]
        remember_rate = 1 - forget_rate
        num_remember = int(remember_rate * n)
        num_start = int((1 - remember_rate) * n)
        
        _, ind_1_sorted = torch.sort(loss_1)
        ind_1_update = ind_1_sorted[:num_remember]
        
        _, ind_2_sorted = torch.sort(loss_2)
        ind_2_update = ind_2_sorted[:num_remember]
    
    
    pure_ratio_1 = torch.sum(noise_or_not[ind[ind_1_sorted[:num_remember]]])/float(num_remember)
    pure_ratio_2 = torch.sum(noise_or_not[ind[ind_2_sorted[:num_remember]]])/float(num_remember)

    # do not exchange: spl-mse

    model1.zero_grad()
    model2.zero_grad()
    y_1 = model1(x[ind_1_update, :, :, :])
    y_2 = model2(x[ind_2_update, :, :, :])
    loss_1_update = F.cross_entropy(y_1, t[ind_1_update])
    loss_2_update = F.cross_entropy(y_2, t[ind_2_update])
    return torch.sum(loss_1_update) / num_remember, torch.sum(loss_2_update) / num_remember, pure_ratio_1, pure_ratio_2


def loss_bootstrapping(x, model1, model2, t, forget_rate, ind, noise_or_not, which_model_grad = 1, num_classes=10):
    noise_or_not = torch.Tensor(noise_or_not).cuda()
    ind = torch.Tensor(ind).cuda().long()
    n = t.shape[0]
    remember_rate = 1 - forget_rate
    num_remember = int(remember_rate * n)
    
    y_1 = model1(x)
    outputs1 = F.softmax(y_1, dim=1)
    _, pred1 = torch.max(outputs1.data, 1)

    loss_1_update = 0.8 * F.cross_entropy(y_1, t) + 0.2 * F.cross_entropy(y_1, pred1)

    y_2 = model2(x)
    outputs2 = F.softmax(y_2, dim=1)
    _, pred2 = torch.max(outputs2.data, 1)

    # loss 2 does not affect bootstrapping method, here is only for main.py to be consistent
    loss_2_update = 0.8 * F.cross_entropy(y_2, t) + 0.2 * F.cross_entropy(y_2, pred2)

    # pure ratio does not exist in booststrapping method, here pure ratio is meaningless, only for main.py to be consistent.
    pure_ratio_1 = torch.sum(noise_or_not[ind[:num_remember]])/float(num_remember)
    pure_ratio_2 = torch.sum(noise_or_not[ind[:num_remember]])/float(num_remember)

    return loss_1_update, loss_2_update, pure_ratio_1, pure_ratio_2



def loss_CO_DoubleRobust(x, model1, model2, t, forget_rate, ind, noise_or_not, which_model_grad = 1):
    # you should assert the model1 and model2 are added hooks
    noise_or_not = torch.Tensor(noise_or_not).cuda()
    ind = torch.Tensor(ind).cuda().long()
    add_hooks(model1)
    add_hooks(model2)

    y_1 = model1(x)
    y_2 = model2(x)

    loss_1 = F.cross_entropy(y_1, t, reduction='none')
    loss_1_bp = torch.sum(loss_1)
    loss_1_bp.backward()
    loss_1 = F.mse_loss(y_1, t)

    loss_2 = F.cross_entropy(y_2, t, reduction='none')
    loss_2_bp = torch.sum(loss_2)
    loss_2_bp.backward()
    loss_2 = F.mse_loss(y_2, t)
    # get individual gradient
    with torch.no_grad():
        n = y_1.shape[0]
        remember_rate = 1 - forget_rate
        num_remember = int(remember_rate * n)
        num_start = int((1 - remember_rate) * n)

        computed_grad_norm_1 = torch.zeros(n).cuda()
        for p in model1.parameters():
            grad_layer = torch.flatten(p.grad_sample.detach(), start_dim=1).squeeze()
            grad_layer = (grad_layer - grad_layer.mean(dim=0))
            del p.grad_sample
            computed_grad_norm_1 = computed_grad_norm_1 + torch.sum(grad_layer ** 2, dim=1)


        _, ind_1_sorted_gd = torch.sort(computed_grad_norm_1)
        _, ind_1_sorted_loss = torch.sort(loss_1)
        ind_1_sorted_gd = ind_1_sorted_gd[:num_remember]
        ind_1_sorted_loss = ind_1_sorted_loss[:num_remember]
        # print(ind_1_sorted_loss.shape)
        # print(ind_1_sorted_gd.shape)
        # print(num_remember)
        ind_1_update = torch.from_numpy(np.intersect1d(ind_1_sorted_gd.cpu().numpy(), ind_1_sorted_loss.cpu().numpy()))
        # print(ind_1_update.shape)
        computed_grad_norm_2 = torch.zeros(n).cuda()
        for p in model2.parameters():
            grad_layer = torch.flatten(p.grad_sample.detach(), start_dim=1).squeeze()
            grad_layer = (grad_layer - grad_layer.mean(dim=0))
            del p.grad_sample
            computed_grad_norm_2 = computed_grad_norm_2 + torch.sum(grad_layer ** 2, dim=1)

        _, ind_2_sorted_gd = torch.sort(computed_grad_norm_2)
        _, ind_2_sorted_loss = torch.sort(loss_2)
        ind_2_sorted_gd = ind_2_sorted_gd[:num_remember]
        ind_2_sorted_loss = ind_2_sorted_loss[:num_remember]
        ind_2_update = torch.from_numpy(np.intersect1d(ind_2_sorted_gd.cpu().numpy(), ind_2_sorted_loss.cpu().numpy()))

    pure_ratio_1 = torch.sum(noise_or_not[ind_1_update]) / float(ind_1_update.shape[0])
    pure_ratio_2 = torch.sum(noise_or_not[ind_2_update]) / float(ind_1_update.shape[0])

    remove_hooks(model1)
    remove_hooks(model2)
    # exchange

    model1.zero_grad()
    model2.zero_grad()
    y_1 = model1(x[ind_2_update, :, :, :])
    y_2 = model2(x[ind_1_update, :, :, :])
    loss_1_update = F.cross_entropy(y_1, t[ind_2_update])
    loss_2_update = F.cross_entropy(y_2, t[ind_1_update])
    return torch.sum(loss_1_update) / num_remember, torch.sum(loss_2_update) / num_remember, pure_ratio_1, pure_ratio_2
